import torch

from womac.config import MFSConfig, MFSMethod, ReferencePoolConfig


def _build_base_reference_pool(
    X: torch.Tensor,
    reference_pool_config: ReferencePoolConfig,
    cols_without_min_resp: torch.Tensor,
    device: torch.device = torch.device("cpu"),
) -> torch.Tensor:
    """
    Create initial reference pool mask for tasks and experts.

    Exclusions applied for:
        1. Insufficient responses per task/expert.
        2. Held-out expert column if jackknife_col is True.
        3. NaN entries in held-out rows if drop_row_nans is True.

    Args:
        X (torch.Tensor): Data matrix shape (m, n) with NaNs for missing.
        reference_pool_config (ReferencePoolConfig): Flags controlling jackknife and NaN rules.
        cols_without_min_resp (torch.Tensor): bool mask of shape (n,) marking insufficient experts.
        device (torch.device): Device for mask tensor allocation.

    Returns:
        torch.Tensor: Bool mask of shape (dim1, dim2, n), where
            dim1 = m if jackknife_row else 1,
            dim2 = n if jackknife_col else 1.
    """
    m, n = X.shape
    nan_locations = torch.where(torch.isnan(X))
    dim1 = m if reference_pool_config.jackknife_row else 1
    dim2 = n if reference_pool_config.jackknife_col else 1
    reference_pool = torch.ones((dim1, dim2, n), device=device, dtype=torch.bool)

    # 1. Exclude rows/cols with insufficient responses
    reference_pool[:, :, cols_without_min_resp] = False

    # 2. if jackknife_col, exclude all expert j at held-out col j
    if reference_pool_config.jackknife_col:
        reference_pool[:, torch.arange(n), torch.arange(n)] = False

    # 3. if drop_row_nans, exclude NaNs in the held-out row
    if reference_pool_config.drop_row_nans:
        reference_pool[nan_locations[0], :, nan_locations[1]] = False

    return reference_pool


def _marginal_feature_screening(
    base_pool: torch.Tensor,
    errors: torch.Tensor,
    mfs_config: MFSConfig,
) -> torch.Tensor:
    """
    Apply marginal feature screening to refine reference pool.

    Filters base_pool entries based on error thresholds or percentiles.

    Args:
        base_pool (torch.Tensor): Bool mask shape (dim1, dim2, n) to update with MFS.
        errors (torch.Tensor): Error scores shape (dim1, n).
        mfs_config (MFSConfig): Configuration for screening method and cutoff.

    Returns:
        torch.Tensor: Bool mask shape (dim1, dim2, n) after screening.
        dim1 = m if jackknife_row else 1,
        dim2 = n if jackknife_col else 1.

    Raises:
        ValueError: For invalid mfs_method or if any (i,j) has no valid references.
    """
    if mfs_config.mfs_method == MFSMethod.THRESHOLD:
        threshold_mask = errors <= mfs_config.mfs_cutoff  # (dim1, n)
        reference_pool = base_pool & threshold_mask.unsqueeze(1)  # (dim1, dim2, n)

    elif mfs_config.mfs_method == MFSMethod.PERCENTILE or mfs_config.mfs_method == MFSMethod.AUTO:
        errors_expanded = errors.unsqueeze(1).repeat(1, base_pool.shape[1], 1)  # (dim1, dim2, n)
        errors_expanded[~base_pool] = torch.nan  # (dim1, dim2, n)

        threshold_mask = torch.nanquantile(
            errors_expanded,
            mfs_config.mfs_cutoff / 100.0,
            dim=-1,
            keepdim=True,
        )  # shape (dim1, dim2, 1)
        reference_pool = base_pool & (errors_expanded <= threshold_mask)  # (dim1, dim2, 1)
    elif mfs_config.mfs_method is MFSMethod.NONE:
        reference_pool = base_pool
    else:
        raise ValueError(f"Invalid mfs_config.mfs_method: {mfs_config.mfs_method}")

    if torch.any(torch.sum(reference_pool, dim=-1) == 0):
        raise ValueError("Some indices have no reference pool. Try raising the mfs_cutoff.")

    return reference_pool



import argparse
import logging

import numpy as np
import pandas as pd
from omegaconf import OmegaConf

from womac.config import WomacConfig
from womac.womac import Womac

logger = logging.getLogger(__name__)


def main():
    """Command-line interface for running WOMAC."""
    # configure logging for CLI output
    logging.basicConfig(
        format="%(asctime)s %(levelname)s: %(message)s",
        datefmt="%H:%M:%S",
        level=logging.INFO,
    )
    parser = argparse.ArgumentParser(description="Run WOMAC scoring from the command line")
    parser.add_argument("--predictions", "-p", required=True, help="Path to predictions file (CSV or .npy)")
    parser.add_argument("--targets", "-t", required=True, help="Path to targets file (CSV or .npy)")
    parser.add_argument("--config", "-c", required=True, help="Path to YAML configuration file")
    parser.add_argument("--output", "-o", default=None, help="Optional path to save leaderboard CSV")
    parser.add_argument("--top_k", "-k", type=int, default=25, help="Log the top K experts indices.")
    args = parser.parse_args()

    def load_array(path):
        if path.endswith(".csv"):
            return pd.read_csv(path).to_numpy()  # expects first row as header
        elif path.endswith(".npy"):
            return np.load(path)
        else:
            raise ValueError(f"Unsupported file format: {path}")

    X = load_array(args.predictions)
    Y = load_array(args.targets).squeeze()

    # Load and validate config via OmegaConf
    raw_cfg = OmegaConf.load(args.config)
    womac_keys = ["min_responses_per_task", "min_responses_per_expert", "device", "seed"]
    womac_init = {key: raw_cfg.pop(key, None) for key in womac_keys}
    schema = OmegaConf.structured(WomacConfig)
    cfg = OmegaConf.merge(schema, raw_cfg)
    wcfg: WomacConfig = OmegaConf.to_object(cfg)

    # Run WOMAC scoring
    model = Womac(**womac_init)
    result = model.score_competition(X, Y, wcfg)

    # Log top-k or full leaderboard
    logger.info("\n%s", result.leaderboard(args.top_k))

    # Optionally save results
    if args.output:
        df = pd.DataFrame(
            {
                "rank": list(range(1, len(result.ranked_indices) + 1)),
                "expert_index": result.ranked_indices.tolist(),
                "score": result.ranked_scores.tolist(),
            }
        )
        df.to_csv(args.output, index=False)



from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import Optional

import numpy as np
import torch

TensorLike = torch.Tensor | np.ndarray

###############################################################################
# Enumerations for configuration options
###############################################################################


# missing data strategies
class MissingDataStrategy(Enum):
    IGNORE_NANS = "ignore_nans"
    IMPUTE_OUTCOME_MEAN = "impute_outcome_mean"
    IMPUTE_COLUMN_MEAN = "impute_column_mean"


# methods for marginal feature screening
class MFSMethod(Enum):
    AUTO = "auto"
    THRESHOLD = "threshold"
    PERCENTILE = "percentile"
    NONE = "none"


# Whether to aggregate or separate features
class FeaturizeAs(Enum):
    AGGREGATE = "aggregate"
    SEPARATE = "separate"


# Fit Methods
class FitMethod(Enum):
    UNIFORM = "uniform"
    WEIGHTED = "weighted"


class Target(Enum):
    BINARY = "binary"
    CONTINUOUS = "continuous"


###############################################################################
# Configuration dataclasses
###############################################################################


@dataclass
class MFSConfig:
    mfs_method: MFSMethod = MFSMethod.PERCENTILE
    mfs_cutoff: float = 15
    start_sweep: Optional[float] = 1
    end_sweep: Optional[float] = 100
    step_size: Optional[float] = 1

    def __post_init__(self):
        if self.mfs_cutoff < 0:
            raise ValueError("mfs_cutoff must be positive.")
        if self.mfs_method == MFSMethod.PERCENTILE and (self.mfs_cutoff > 100 or self.mfs_cutoff < 0):
            raise ValueError("mfs_cutoff for PERCENTILE method must be <= 100.")
        if self.start_sweep <= 0:
            raise ValueError("start_sweep must be greater than 0.")
        if self.end_sweep <= self.start_sweep or self.end_sweep > 100:
            raise ValueError("end_sweep must be greater than start_sweep and less than or equal to 100.")
        if self.step_size <= 0:
            raise ValueError("step_size must be at greater than 0.")


@dataclass
class ReferencePoolConfig:
    jackknife_row: bool = True  # (m, 1, n)
    jackknife_col: bool = True  # (1, n, n)
    drop_row_nans: bool = True
    missing_data_strategy: MissingDataStrategy = MissingDataStrategy.IMPUTE_OUTCOME_MEAN
    mfs_config: Optional[MFSConfig] = field(default_factory=MFSConfig)

    def __post_init__(self):
        if self.drop_row_nans and not self.jackknife_row:
            raise ValueError("Cannot drop row NaNs without jackknife_row enabled.")
        if self.jackknife_row and self.mfs_config is None:
            raise ValueError(
                "jackknife_row cannot be enabled without a marginal feature screening method; "
                "otherwise all jackknifed rows will have the same reference pool.",
            )


@dataclass
class FeatureConfig:
    missing_data_strategy: MissingDataStrategy = MissingDataStrategy.IMPUTE_OUTCOME_MEAN
    missingness_features: Optional[FeaturizeAs] = None
    pred_features: FeaturizeAs = FeaturizeAs.AGGREGATE
    logit_feats: bool = False  # Only used for binary outcomes
    score_as_probs: bool = False  # Only relevant for binary outcomes and when logit_feats is True (ignored otherwise)

    def __post_init__(self):
        if self.missing_data_strategy == MissingDataStrategy.IGNORE_NANS:
            raise ValueError("missing_data_strategy for FeatureConfig cannot be IGNORE_NANS!")


@dataclass
class ModelConfig:
    fit_method: FitMethod = FitMethod.UNIFORM
    epochs: int = 500
    lr: float = 0.1
    l2: float = 0.1
    log_interval: int = 50

    def kwargs_as_dict(self) -> dict:
        """Return model config as dict (without fit_method)."""
        d = asdict(self)
        d.pop("fit_method", None)
        return d


@dataclass
class WomacConfig:
    target: Target
    reference_pool_config: ReferencePoolConfig = field(default_factory=ReferencePoolConfig)
    feature_config: FeatureConfig = field(default_factory=FeatureConfig)
    model_config: ModelConfig = field(default_factory=ModelConfig)

    def __post_init__(self):
        if self.feature_config.score_as_probs and self.target == Target.CONTINUOUS:
            raise ValueError("score_as_probs requires target to be binary.")
        if self.feature_config.logit_feats and self.target == Target.CONTINUOUS:
            raise ValueError("logit_feats cannot be enabled for continuous outcomes.")

    @classmethod
    def binary_outcome(cls) -> "WomacConfig":
        """Configure for binary outcomes with probabilistic predictions.

        Returns:
            WomacConfig: Configuration for logistic regression on binary outcomes with logit transformation enabled.
        """
        return cls(
            target=Target.BINARY,
            feature_config=FeatureConfig(
                logit_feats=True,
                score_as_probs=True,
            ),
        )

    @classmethod
    def continuous_outcome(cls) -> "WomacConfig":
        """Configure for continuous outcomes with linear predictions.

        Returns:
            WomacConfig: Configuration for linear regression on continuous outcomes.
        """
        return cls(
            target=Target.CONTINUOUS,
        )


###############################################################################
# Data container for WOMAC algorithm
###############################################################################


@dataclass(frozen=True, slots=True)
class WomacData:
    """Container for WOMAC algorithm input data.

    Attributes:
        train_preds (Optional[torch.Tensor]): Prediction tensor for training of shape (dim1, dim2, k, l),
            where dim1 = m if jackknife_row else 1, dim2 = n if jackknife_col else 1,
            and k ∈ {m-1, m} (depending on jackknife_row) and l ∈ {1, n} (depending on pred_features).
        train_missingness (Optional[torch.Tensor]): Missingness indicator tensor for training of same shape as train_preds.
        train_labels (torch.Tensor): Labels tensor for training of shape (dim1, 1, k).
        test_preds (torch.Tensor): Prediction tensor for testing of shape (m, dim2, 1, l).
        test_missingness (Optional[torch.Tensor]): Missingness indicator tensor for testing of same shape as test_preds.
        test_labels (Optional[torch.Tensor]): Labels tensor for testing of shape (m, dim2, 1) if provided.
    """

    train_preds: Optional[torch.Tensor]
    train_missingness: Optional[torch.Tensor]
    train_labels: torch.Tensor
    test_preds: torch.Tensor
    test_missingness: Optional[torch.Tensor]
    test_labels: Optional[torch.Tensor] = None

    @property
    def batch_dim1(self) -> int:
        """Size of the first batch dimension (m). This is m if jackknife_row
        is True, otherwise 1.
        """
        if self.train_preds is not None:
            return self.train_preds.shape[0]
        raise ValueError("Invalid training predictions provided.")

    @property
    def batch_dim2(self) -> int:
        """Size of the second batch dimension (n). This is n if jackknife_col
        is True, otherwise 1.
        """
        if self.train_preds is not None:
            return self.train_preds.shape[1]
        raise ValueError("Invalid training predictions provided.")

    @property
    def data_dim(self) -> int:
        """Size of the data dimension (k1).

        This is the number of tasks to predict over at a time.
        During training, this is m-1 if jackknife_row is True, otherwise m.
        During testing, this is 1.
        """
        if self.train_preds is not None:
            return self.train_preds.shape[2]
        raise ValueError("Invalid training predictions provided.")

    @property
    def feature_dim(self) -> int:
        """Size of the feature dimension (pred plus missingness features).

        This is:
        1 if pred_features is AGGREGATE and missingness_features is None,
        2 if pred_features is AGGREGATE and missingness_features is AGGREGATE,
        n if pred_features is SEPARATE and missingness_features is None,
        n + 1 if pred_features is SEPARATE and missingness_features is AGGREGATE OR
              if pred_features is AGGREGATE and missingness_features is SEPARATE.
        2 * n if pred_features is SEPARATE and missingness_features is SEPARATE.
        """
        train_data = self.train_data
        if train_data is not None:
            return train_data.shape[3]
        raise ValueError("Invalid training predictions provided.")

    @property
    def train_data(self) -> torch.Tensor:
        """Construct combined training data tensor by concatenating predictions and missingness
        features.

        If train_missingness is provided, returns a tensor of shape (dim1, dim2, k, l1 + l2) by
        concatenating along dim=3; otherwise returns train_preds unchanged.
        """
        if self.train_missingness is None:
            return self.train_preds
        else:
            return torch.cat([self.train_preds, self.train_missingness], dim=3)

    @property
    def test_data(self) -> torch.Tensor:
        """Construct combined test data tensor by concatenating predictions and missingness
        features.

        If test_missingness is provided, returns a tensor of shape (m, n, 1, l1 + l2) by
        concatenating along dim=3; otherwise returns test_preds unchanged.
        """
        if self.test_missingness is None:
            return self.test_preds
        else:
            return torch.cat([self.test_preds, self.test_missingness], dim=3)

    def __post_init__(self):
        if self.train_preds is None and self.train_missingness is not None:
            raise ValueError("train_missingness cannot be provided without train_preds.")
        if self.test_preds is None and self.test_missingness is not None:
            raise ValueError("test_missingness cannot be provided without test_preds.")


###############################################################################
# Results dataclasses
###############################################################################


@dataclass
class ModelResult:
    coef_: torch.Tensor
    train_losses: list[TensorLike]
    bias_: Optional[torch.Tensor] = None
    val_losses: Optional[list[TensorLike]] = None  # This is ONLY val loss; ignores regularization loss.


@dataclass
class WomacResult:
    """Container for WOMAC algorithm output.

    Attributes:
        womac_ranked_indices (torch.Tensor): Expert indices sorted by WOMAC performance, shape (n experts,).
        womac_ranked_scores (torch.Tensor): Corresponding scores for womac_ranked_indices, shape (n experts,).
        womac_scores (torch.Tensor): Raw WOMAC scores per expert, shape (n experts,).
        outcome_ranked_indices (torch.Tensor): Expert indices sorted by outcome scores, shape (n experts,).
        outcome_ranked_scores (torch.Tensor): Corresponding scores for outcome_ranked_indices, shape (n experts,).
        outcome_scores (torch.Tensor): Raw outcome scores per expert, shape (n experts,).
        reference_matrix (torch.Tensor): Reference solutions per expert, shape (m tasks, dim2).
        dim2 is n if jackknife_col is True, otherwise 1.
    """

    womac_ranked_indices: torch.Tensor
    womac_ranked_scores: torch.Tensor
    womac_scores: torch.Tensor
    outcome_ranked_indices: torch.Tensor
    outcome_ranked_scores: torch.Tensor
    outcome_scores: torch.Tensor
    reference_matrix: torch.Tensor
    fitted_model: ModelResult

    @property
    def winner(self, by_womac: bool = True) -> int:
        """Get the index of the best-performing expert.

        Args:
            by_womac (bool): Whether to rank by WOMAC scores or outcome scores.

        Returns:
            int: Expert index with highest performance.
        """
        ranked_indices = self.womac_ranked_indices if by_womac else self.outcome_ranked_indices
        return int(ranked_indices[0].item())

    @property
    def reference_solution(self) -> torch.Tensor:
        """Compute the averaged reference solution across experts."""
        return torch.mean(self.reference_matrix, dim=1)  # (m,)

    def reference_for_expert(self, index: int) -> torch.Tensor:
        """Retrieve the reference solution for a specific expert based
        on their original index.

        Args:
            index (int): Expert index (0 <= index < num_experts).

        Returns:
            torch.Tensor: Reference solution vector for the specified expert, shape (m,).

        Raises:
            IndexError: If index is out of bounds.
        """
        if index < 0 or index >= self.womac_scores.shape[0]:
            raise IndexError("Index out of bounds.")
        if self.reference_matrix.ndim == 1:
            return self.reference_matrix
        return self.reference_matrix[:, index]

    def top_k(self, k: int, by_womac: bool = True) -> torch.Tensor:
        """Return the indices of the top k experts.

        Args:
            k (int): Number of top experts to retrieve.
            by_womac (bool): Whether to rank by WOMAC scores or outcome scores.

        Returns:
            torch.Tensor: Tensor of top k expert indices, shape (k,).
        """
        ranked_indices = self.womac_ranked_indices if by_womac else self.outcome_ranked_indices
        return ranked_indices[:k]

    def score_of(self, index: int, by_womac: bool = True) -> float:
        """Get the score for a specific expert based on their
        original index.

        Args:
            index (int): Rank index in ranked_scores (0-based).
            by_womac (bool): Whether to use WOMAC scores or outcome scores.

        Returns:
            float: Score for the expert at the specified position.

        Raises:
            IndexError: If index is out of bounds.
        """
        scores = self.womac_scores if by_womac else self.outcome_scores
        if index < 0 or index >= scores.shape[0]:
            raise IndexError(f"Index out of bounds; should be between 0 and {scores.shape[0] - 1}.")
        return float(scores[index].item())

    def rank_of(self, index: int, by_womac: bool = True) -> int:
        """Get the performance rank for a specific expert based
        on their original index.

        Args:
            index (int): Expert index to query.
            by_womac (bool): Whether to rank by WOMAC scores or outcome scores.

        Returns:
            int: 1-based rank of the expert.

        Raises:
            IndexError: If index is out of bounds.
        """
        ranked_indices = self.womac_ranked_indices if by_womac else self.outcome_ranked_indices
        if index < 0 or index >= len(ranked_indices):
            raise IndexError(f"Index out of bounds; should be between 0 and {len(ranked_indices) - 1}.")
        return int((ranked_indices == index).nonzero(as_tuple=True)[0].item() + 1)

    def leaderboard(self, limit: Optional[int] = None, by_womac: bool = True) -> str:
        """Generate a human-readable leaderboard of experts showing both WOMAC and outcome scores,
        sorted by the selected metric.

        Args:
            limit (Optional[int]): Maximum number of top experts to include. If None, includes all.
            by_womac (bool): Whether to sort by WOMAC scores or outcome scores.

        Returns:
            str: Multiline string table with columns Rank, Expert Index, WOMAC MSE, and Outcome MSE.
        """
        # Choose the ranking order
        order = self.womac_ranked_indices if by_womac else self.outcome_ranked_indices

        # Determine how many entries to show
        if limit is None:
            limit = len(order)

        # Header with both columns
        lines = ["Rank | Expert Index | WOMAC MSE  | Outcome MSE"]
        for rank, idx in enumerate(order[:limit], start=1):
            idx_int = int(idx.item())
            womac_score = self.womac_scores[idx_int].item()
            outcome_score = self.outcome_scores[idx_int].item()
            lines.append(f"{rank:4} | {idx_int:12} | {womac_score:10.7f} | {outcome_score:11.7f}")

        return "\n".join(lines)



import logging
from typing import Optional

import torch
import torch.nn as nn
from tqdm import tqdm

from womac.config import ModelResult, Target

logger = logging.getLogger(__name__)


class AverageModel:
    """Model that computes predictions by averaging selected experts with uniform weights.

    Attributes:
        coef_ (None or torch.Tensor): Coefficients of the model, initialized as None.
        target (Target): The target object passed during initialization.
        model_result (None or ModelResult): The result of the model computation, initialized as None.
        logit_feats (bool): Indicates whether the inputs are logit-transformed.
    """

    def __init__(self, target: Target, logit_feats: bool):
        """
        Initialize the AverageModel class.

        Args:
            target (Target): The target object that defines the model's target variable.
            logit_feats (bool): A flag indicating whether the target variable is logit-transformed.

        Attributes:
            coef_ (None or torch.Tensor): Coefficients of the model, initialized as None.
            target (Target): The target object passed during initialization.
            model_result (None or ModelResult): The result of the model computation, initialized as None.
            logit_feats (bool): Indicates whether the inputs are logit-transformed (ONLY for binary targets).
        """
        self.coef_ = None
        self.target = target
        self.model_result = None
        self.logit_feats = logit_feats
        if self.logit_feats and self.target != Target.BINARY:
            raise ValueError("logit_feats can only be True for binary targets.")

        logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")

    def fit(
        self,
        mask: torch.Tensor,
        val_data: Optional[torch.Tensor] = None,
        val_labels: Optional[torch.Tensor] = None,
    ) -> None:
        """Set coefficients for averaging based on a one-hot mask.
        If the target is BINARY:
            - If logit_feats is True, model will average over logits (geometric pooling).
            Val loss computed as cross-entropy loss (after a sigmoid) against a binary label.
            - If logit_feats is False, model will average over probabilities (arithmetic pooling).
            Val loss computed directly as cross-entropy loss against a binary label.
        If the target is CONTINUOUS, model will average over continuous values (arithmetic pooling).
        Val loss computed as MSE loss against a continuous label.

        Args:
            mask (Tensor): One-hot tensor of shape (dim1, dim2, n) where mask[i, j, k] = 1
                indicates that for task i and expert j, expert k is included in the average.
                dim1 is m if jackknife_row is True, otherwise 1;
                dim2 is n if jackknife_col is True, otherwise 1.
            val_data (Optional[Tensor]): Validation data tensor of shape [m, dim2, 1, l].
            val_targets (Optional[Tensor]): Validation targets tensor of shape [m, dim2, 1].

        Returns:
            None. Sets self.coef_ to a Tensor of shape (m, n, n) containing normalized weights.
        """
        val_losses = []

        self.coef_ = mask / torch.sum(mask, dim=2, keepdims=True)  # (dim1, dim2, n)

        if val_data is not None and val_labels is not None:
            if self.target == Target.BINARY and self.logit_feats:
                criterion = nn.BCEWithLogitsLoss()
            elif self.target == Target.BINARY and not self.logit_feats:
                criterion = nn.BCELoss()
            else:
                criterion = nn.MSELoss()
            val_preds = self.predict(val_data)
            val_loss = criterion(val_preds, val_labels)
            val_losses.append(val_loss.detach().cpu().float())

        self.model_result = ModelResult(
            coef_=self.coef_.cpu(),
            train_losses=[],
            bias_=None,
            val_losses=val_losses,
        )

    def predict(self, X: torch.Tensor) -> torch.Tensor:
        """Predict averaged outputs using mask coefficients.

        Args:
            X (Tensor): Input tensor of shape (m, 1, n), where
                m: number of tasks,
                n: number of experts,

        Returns:
            Tensor of shape (m, dim2, 1) containing averaged predictions.
        """
        # (dim1, dim2, n) @ (m, 1, n).T -> (m, dim2, 1)
        return torch.matmul(self.coef_, X.transpose(1, 2))  # (m, dim2, 1)

    def predict_proba(self, X: torch.Tensor) -> torch.Tensor:
        """Compute probabilities for binary classification.

        Args:
            X (Tensor): Input tensor of shape [m, dim2, 1, l].

        Returns:
            Tensor of shape [m, dim2, 1]: Probabilities for binary classification.

        Raises:
            ValueError: If the target is not binary or if logit transformation was not applied.
        """
        if not self.logit_feats:
            raise ValueError("predict_proba is only available when inputs are logit transformed (logit_feats=True).")
        logits = self.predict(X)
        return torch.sigmoid(logits)


class LinearModel(nn.Module):
    """Parallel linear or logistic regression models across batch dimensions.

    Implements dim1 x dim2 separate regressions (or classifiers) in parallel.
    dim1 is m if jackknife_row is True, otherwise 1;
    dim2 is n if jackknife_col is True, otherwise 1.

    Attributes:
        W (Parameter): Weight tensor of shape (dim1, dim2, input_dim, 1).
        b (Parameter): Bias tensor of shape (m, n, 1).
        target (Target): Whether the target labels/outcomes are BINARY or CONTINUOUS.
        logit_feats (bool): Whether the inputs are logit-transformed (ONLY for binary targets).
        model_result (None or ModelResult): The result of the model computation, initialized as None.
    """

    def __init__(
        self,
        dim1: int,
        dim2: int,
        input_dim: int,
        target: Target,
        logit_feats: bool,
        device: str = "cpu",
    ):
        """Initialize batch-wise linear or logistic models.

        Args:
            dim1 (int): Number of tasks (first batch dimension, m).
            dim2 (int): Number of experts (second batch dimension, n).
            input_dim (int): Number of input features (l).
            target (Target): Whether the target labels/outcomes are BINARY or CONTINUOUS.
            logit_feats (bool): Whether the inputs are logit-transformed (ONLY for binary targets).
            device (str): Device to allocate parameters.
        """
        super().__init__()
        self.W = nn.Parameter(torch.randn(dim1, dim2, input_dim, 1, device=device))  # (dim1, dim2, l, 1)
        self.b = nn.Parameter(torch.zeros(dim1, dim2, 1, device=device))  # (dim1, dim2, 1)
        self.target = target
        self.logit_feats = logit_feats
        if self.logit_feats and self.target != Target.BINARY:
            raise ValueError("logit_feats can only be True for binary targets.")
        self.model_result = None  # Will be set after training
        logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """Compute forward pass for parallel linear models.

        Args:
            X (Tensor): Input tensor of shape [dim1, dim2, k, l] during training, where
                dim1: batch dim (tasks),
                dim2: batch dim (experts),
                k: number of samples to predict,
                l: features
                During testing, shape is [m, dim2, 1, l] where m is number of tasks.
        Returns:
            Tensor of shape [m, n, k]: Raw outputs (logits).
        """

        Z = torch.matmul(X, self.W).squeeze(-1)  # [dim1, dim2, k] or [m, dim2, 1]
        Z = Z + self.b  # [dim1, dim2, k] or [m, dim2, 1]
        return Z  # [dim1, dim2, k] (train) or [m, dim2, 1] (test)

    def train(
        self,
        train_data: torch.Tensor,
        train_labels: torch.Tensor,
        epochs: int,
        lr: float,
        l2: float,
        log_interval: int,
        val_data: Optional[torch.Tensor] = None,
        val_labels: Optional[torch.Tensor] = None,
    ) -> None:
        """Train all parallel models using SGD with optional L2 regularization.

        Args:
            train_data (Tensor): Training data tensor of shape [dim1, dim2, k, l].
            train_labels (Tensor): Targets tensor of shape [dim1, 1, k].
            epochs (int): Number of training epochs.
            lr (float): Learning rate.
            l2 (float): L2 regularization coefficient on weights. Defaults to 0.0.
            log_interval (int): Interval for logging training progress.
            val_data (Optional[Tensor]): Validation data tensor of shape [m, dim2, 1, l].
            val_targets (Optional[Tensor]): Validation targets tensor of shape [m, dim2, 1].
        """

        y_b = train_labels.expand(-1, train_data.shape[1], -1).float()  # [dim1, dim2, k]
        if self.target == Target.BINARY:
            criterion = nn.BCEWithLogitsLoss()
        else:
            criterion = nn.MSELoss()
        optimizer = torch.optim.SGD(self.parameters(), lr=lr)
        train_losses, val_losses = [], []

        logger.info(
            f"Training {self.__class__.__name__} model with {epochs} epochs and learning rate {lr}, l2 penalty={l2}."
        )
        for epoch in tqdm(range(epochs), desc="Training Progress"):
            optimizer.zero_grad()
            preds = self.forward(train_data)  # [dim1, dim2, k]
            loss = criterion(preds, y_b)
            if l2 > 0:
                loss = loss + l2 * self.W.pow(2).sum()
            loss.backward()
            optimizer.step()
            train_losses.append(loss.detach().cpu().float())

            if val_data is not None and val_labels is not None:
                with torch.no_grad():
                    val_preds = self.forward(val_data)
                    val_loss = criterion(val_preds, val_labels)
                    reg_loss = (l2 * self.W.pow(2).sum()).item() if l2 > 0 else 0
                val_losses.append(val_loss.detach().cpu().float())

            if (epoch + 1) % log_interval == 0 or epoch == 0:
                val_loss_msg = (
                    f"| Val Loss: {val_loss.item():.6f} | Reg Loss: {reg_loss:.6f} | Total Val Loss: {(val_loss + reg_loss).item():.6f}"
                    if val_data is not None
                    else ""
                )
                tqdm.write(f"Epoch {epoch + 1}/{epochs}, Train Loss: {loss.item():.6f}{val_loss_msg}")

        self.model_result = ModelResult(
            coef_=self.W.detach().cpu(),
            train_losses=train_losses,
            bias_=self.b.detach().cpu(),
            val_losses=val_losses,
        )

    def predict(self, X: torch.Tensor) -> torch.Tensor:
        """Compute predictions without gradient tracking.

        Args:
            X (Tensor): Input tensor of shape [m, dim2, 1, l].

        Returns:
            Tensor of shape [m, dim2, 1]: Predictions (values or probabilities).
        """
        with torch.no_grad():
            return self.forward(X)  # [m, dim2, 1]

    def predict_proba(self, X: torch.Tensor) -> torch.Tensor:
        """Compute probabilities for binary classification.

        Args:
            X (Tensor): Input tensor of shape [m, dim2, 1, l].

        Returns:
            Tensor of shape [m, dim2, 1]: Probabilities for binary classification.
        """
        if not self.logit_feats:
            raise ValueError("predict_proba is only available when inputs are logit transformed (logit_feats=True).")
        logits = self.predict(X)
        return torch.sigmoid(logits)



import torch

from womac.config import FeaturizeAs
from womac.utils import jackknife_expansion


def _featurize_and_center(
    X_train: torch.Tensor,
    X_test: torch.Tensor,
    mask: torch.Tensor,
    pred_features: FeaturizeAs,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Featurize (aggregate predictions or keep separate) and center training and test data.

    Args:
        X_train: Tensor of shape (dim1, k, n). If jackknife_row is True,
            k is m-1 and dim1 is m; otherwise k is m and dim1 is 1.
        X_test: Tensor of shape (m, 1, n), for m tasks and n experts.
        mask: Tensor of shape (dim1, dim2, n). Mask indicating which experts can be used for
          each task. If jackknife_row is True, dim1 is m; otherwise dim1 is 1.
        If jackknife_col is True, dim2 is n; otherwise dim2 is 1.
        pred_features: FeaturizeAs. Specifies AGGREGATE or SEPARATE featurization.

    Returns:
        A tuple of (X_train, X_test):
            X_train: Centered training tensor. Shapes:
                - AGGREGATE: (dim1, dim2, k, 1)
                - SEPARATE:  (dim1, dim2, k, n)
            X_test: Centered test tensor. Shapes:
                - AGGREGATE: (m, dim2, 1, 1)
                - SEPARATE:  (m, dim2, 1, n)
    """

    if pred_features == FeaturizeAs.AGGREGATE:
        average_mask = mask / torch.sum(mask, dim=-1, keepdims=True)  # (dim1, dim2, n)

        # (dim1, dim2, n) @ (dim1, n, k) -> (dim1, dim2, k) -> (dim1, dim2, k, 1)
        X_train = torch.matmul(average_mask, X_train.transpose(1, 2)).unsqueeze(3)
        # (dim1, dim2, 1, n) @ (m, 1, n, 1) -> (m, dim2, 1, 1)
        X_test = torch.matmul(average_mask.unsqueeze(-2), X_test.unsqueeze(-1))

    else:
        # (dim1, 1, k, n) * (dim1, dim2, 1, n) -> (dim1, dim2, k, n)
        X_train = X_train.unsqueeze(1) * mask.unsqueeze(2)
        # (m, 1, 1, n) * (dim1, dim2, 1, n) -> (m, dim2, 1, n)
        X_test = X_test.unsqueeze(1) * mask.unsqueeze(2)

    X_mean = torch.mean(X_train, dim=[2, 3], keepdim=True)  # (dim1, dim2, 1, 1)
    X_std = torch.std(X_train, dim=[2, 3], keepdim=True)  # (dim2, dim2, 1, 1)
    X_train = (X_train - X_mean) / X_std  # (dim1, dim2, k, l)
    X_test = (X_test - X_mean) / X_std  # (m, dim2, 1, l)

    return X_train, X_test


def _missingness_features(
    X: torch.Tensor,
    mask: torch.Tensor,
    missingness_features: FeaturizeAs,
    jackknife_row: bool,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Compute missingness indicator features for training and test data.

    Args:
        X: Tensor of shape (m, n) with input observations that may contain NaNs.
        mask: Tensor of shape (dim1, dim2, n). Mask indicating which experts can be used for
            each task. If jackknife_row is True, dim1 is m; otherwise dim1 is 1. If
            jackknife_col is True, dim2 is n; otherwise dim2 is 1.
        missingness_features: FeaturizeAs enum specifying featurization method
            None or (AGGREGATE or SEPARATE).
        jackknife_row: Boolean flag indicating whether to apply jackknife expansion
            to rows of X, producing k = m-1 training rows if True.

    Returns:
        Tuple of (X_train_nans, X_test_nans):
            X_train_nans: Tensor or None containing training missingness features:
                - AGGREGATE: shape (dim1, dim2, k, 1)
                - SEPARATE: shape (dim1, dim2, k, n).
                - NONE: None.
                If jackknife_row is True, k = m-1; otherwise k = m.
            X_test_nans: Tensor or None containing test missingness features:
                - AGGREGATE: shape (m, dim2, 1, 1).
                - SEPARATE: shape (m, dim2, 1, n).
                - NONE: None.
    """

    if jackknife_row:
        X_transf = jackknife_expansion(X)  # (m, m-1, n)
        train_nan_mask = torch.isnan(X_transf).float()  # (m, m-1, n)
    else:
        X_transf = X.unsqueeze(0)  # (1, m, n)
        train_nan_mask = torch.isnan(X_transf).float()  # (1, m, n)
    test_nan_mask = torch.isnan(X).float().unsqueeze(1)  #  (m, 1, n)

    if missingness_features == FeaturizeAs.AGGREGATE:
        average_mask = mask / torch.sum(mask, dim=-1, keepdim=True)  # (dim1, dim2, n)

        # (dim1, dim2, n) @ (dim1, n, k) -> (dim1, dim2, k) -> (dim1, dim2, k, 1)
        X_train_nans = torch.matmul(average_mask, train_nan_mask.transpose(1, 2)).unsqueeze(3)
        # (dim1, dim2, 1, n) @ (m, 1, n, 1) -> (m, dim2, 1, 1)
        X_test_nans = torch.matmul(average_mask.unsqueeze(-2), test_nan_mask.unsqueeze(-1))

    elif missingness_features == FeaturizeAs.SEPARATE:
        # (dim1, 1, k, n) * (dim1, dim2, 1, n) -> (dim1, dim2, k, n)
        X_train_nans = train_nan_mask.unsqueeze(1) * mask.unsqueeze(2)
        # (m, 1, 1, n) * (dim1, dim2, 1, n) -> (m, dim2, 1, n)
        X_test_nans = test_nan_mask.unsqueeze(1) * mask.unsqueeze(2)

    else:
        X_train_nans = None
        X_test_nans = None

    return X_train_nans, X_test_nans  # (dim1, dim2, k, l), (m, dim2, 1, l) or None, None



from typing import Optional

import numpy as np
import torch

from womac.config import MissingDataStrategy, TensorLike


def _cast_to_tensor(X: TensorLike, device: torch.device = torch.device("cpu")) -> torch.Tensor | None:
    """
    Converts the input data to a torch tensor. Reshapes 1D arrays to 2D with shape (-1, 1).

    Args:
        X (TensorLike): The input data to be converted. It can be either a numpy array,
            a torch tensor, or None.
        device (torch.device): The device on which the resulting tensor should be placed.
            Defaults to CPU.

    Returns:
        torch.Tensor | None: A 2D torch tensor representation of the input data, or None

    Raises:
        TypeError: If the input is not a numpy array, torch tensor, or None.
    """
    if isinstance(X, np.ndarray):
        if X.ndim == 1:
            X = X.reshape(-1, 1)
        return torch.tensor(X, device=device, dtype=torch.float32)
    if isinstance(X, torch.Tensor):
        if X.ndim == 1:
            X = X.view(-1, 1)
        return X.clone().to(device)
    if X is None:
        return None
    raise TypeError(f"Expected input to be a numpy array, torch tensor, or None, but got {type(X)}.")


def _apply_logit_transform(X: torch.Tensor, logit_transform: bool) -> torch.Tensor:
    """
    Applies a logit transformation to input tensor values if requested.

    Args:
        X (torch.Tensor): Tensor containing values strictly between 0 and 1.
        logit_transform (bool): Whether to apply the logit transform.

    Returns:
        torch.Tensor: Logit-transformed tensor when logit_transform is True; otherwise original tensor.

    Raises:
        ValueError: If logit_transform is True and any values in X are <= 0 or >= 1.
    """
    if logit_transform:
        if torch.any((X <= 0) | (X >= 1)):  # scalar bool
            raise ValueError(
                "All entries must be strictly between 0 and 1 for logit transform. Check your data or disable logit transformation."
            )
        return torch.logit(X)
    return X


def _get_indices_without_min_responses(
    X: torch.Tensor,
    min_responses_per_row: int,
    min_responses_per_col: int,
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Compute boolean masks indicating which rows and columns of the input
    tensor have fewer valid (non-NaN) entries than the configured minimum thresholds:
    `self.min_responses_per_task` for rows (tasks) and
    `self.min_responses_per_expert` for columns (experts).

    Args:
        X (torch.Tensor): A 2D tensor of shape (m, n). Missing or invalid responses
            should be encoded as NaN.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: A tuple containing:
            rows_without_min_resp (torch.Tensor): A 1D boolean tensor of length m tasks.
                True at position i indicates that task i has fewer than
                `self.min_responses_per_task` valid responses.
            cols_without_min_resp (torch.Tensor): A 1D boolean tensor of length n experts.
                True at position j indicates that expert j has fewer than
                `self.min_responses_per_expert` valid responses.
    """
    is_not_nan = ~torch.isnan(X)
    rows_without_min_resp = torch.sum(is_not_nan, dim=1) < min_responses_per_row
    cols_without_min_resp = torch.sum(is_not_nan, dim=0) < min_responses_per_col

    return rows_without_min_resp, cols_without_min_resp


######### Jackknife Expansion Function #########


def jackknife_expansion(X: torch.Tensor) -> torch.Tensor:
    """
    Generate jackknife samples by excluding each row in turn.

    Args:
        X (torch.Tensor): A 2D tensor of shape (m, *).

    Returns:
        torch.Tensor: A tensor of shape (m, m-1, *) where each slice along the first
            dimension is X with the corresponding row removed.
    """
    m = X.size(0)
    X_rep = X.unsqueeze(0).expand(m, -1, *X.shape[1:])  # (m, m, *)
    mask = ~torch.eye(m, dtype=torch.bool, device=X.device)  # (m, m)
    return X_rep[mask].view(m, m - 1, *X.shape[1:])


def score_mse(
    X: torch.Tensor,
    Y: torch.Tensor,
    missing_data_strategy: MissingDataStrategy,
    jackknife_row: bool = False,
    apply_logit_transform: bool = False,
) -> torch.Tensor:
    """
    Compute mean squared error (MSE) between predictions and targets for each expert.

    If `jackknife_row` is True, performs leave-one-row-out (jackknife) MSE for each row.
    Otherwise, computes MSE for each expert over all rows.

    Args:
        X (torch.Tensor): Prediction matrix (m, n), may contain NaNs.
        Y (torch.Tensor): Outcome vector or matrix (m,) or (m, 1), may contain NaNs.
        missing_data_strategy (MissingDataStrategy): How to handle NaNs in X.
        jackknife_row (bool): If True, use jackknife; else, compute over all rows.
        apply_logit_transform (bool): Whether to apply logit transformation to X before scoring.

    Returns:
        torch.Tensor: MSE values, shape (m, n) if jackknife, else (1, n).

    Raises:
        ValueError: If NaNs present in errors (due to too many NaNs in Y).
    """
    X, Y = _cast_to_tensor(X, device=X.device), _cast_to_tensor(Y, device=Y.device)
    # Optionally expand for jackknife
    if jackknife_row:
        X = jackknife_expansion(X)  # (m, m-1, n)
        Y = jackknife_expansion(Y)  # (m, m-1, 1)
    else:
        X = X.unsqueeze(0)  # (1, m, n)
        Y = Y.unsqueeze(0)  # (1, m, 1)

    # Impute or ignore NaNs in X_sc (apply_logit_transform=False since scoring directly against targets Y_sc)
    X_sc, _ = impute_data(X, missing_data_strategy, Y=Y)  # (m, m-1, n) or (1, m, n)

    # Apply logit transformation if configured
    X_sc = _apply_logit_transform(X_sc, apply_logit_transform)  # (m, m-1, n) or (1, m, n)

    # Compute mean squared errors
    errors = torch.nanmean((X_sc - Y) ** 2, dim=-2)  # (m, n) (if jackknife_row=True) or (1, n) (if jackknife_row=False)

    # Ensure no NaNs remain
    if torch.isnan(errors).any():
        raise ValueError("NaN values in errors matrix. Reduce `min_responses_per_task` or provide more data.")

    return errors


######### Imputation Functions #########


def impute_column_mean(X: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Imputes missing entries in X with the mean of each column (second-to-last dimension).

    Args:
        X (torch.Tensor): Input tensor of any shape with NaN values along the second-to-last
          dimension.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]:
            - imputed_X (torch.Tensor): Tensor with NaNs replaced by column means.
            - column_mean (torch.Tensor): Tensor of column means with shape (*, 1, n).
    """
    column_mean = torch.nanmean(X, dim=-2, keepdims=True)  # (*, 1, n)
    return torch.where(torch.isnan(X), column_mean, X), column_mean  # (*, k, n)


def impute_outcome_mean(X: torch.Tensor, Y: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Impute missing entries in X using the mean of outcome tensor Y.

    Args:
        X (torch.Tensor): Input tensor with NaN values to be imputed.
        Y (torch.Tensor): Outcome tensor used to compute the fill-in mean.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]:
            - imputed_X (torch.Tensor): Tensor with NaNs replaced by the mean of Y.
            - Y_mean (torch.Tensor): Scalar or tensor containing the mean of Y.
    """
    Y_mean = torch.nanmean(Y, dim=-2, keepdim=True)  # (1, 1)
    imputed = torch.where(torch.isnan(X), Y_mean, X)  # (*, k, n)
    return imputed, Y_mean


def impute_data(
    X: torch.Tensor,
    missing_data_strategy: MissingDataStrategy,
    Y: Optional[torch.Tensor],
) -> tuple[torch.Tensor, Optional[torch.Tensor]]:
    """
    Impute missing data in tensor X according to a specified strategy.

    Args:
        X (torch.Tensor): Input tensor with potential NaN values.
        missing_data_strategy (MissingDataStrategy): Strategy for handling missing data.
        Y (Optional[torch.Tensor]): Optional outcome tensor required by some strategies.

    Returns:
        Tuple[torch.Tensor, Optional[torch.Tensor]]: A tuple containing:
            - imputed_X (torch.Tensor): Tensor with missing values handled.
            - imputation_info (Optional[torch.Tensor]): The tensor used for imputation.

    Raises:
        ValueError: If an invalid missing_data_strategy is provided.
    """
    if missing_data_strategy == MissingDataStrategy.IMPUTE_COLUMN_MEAN:
        return impute_column_mean(X)
    elif missing_data_strategy == MissingDataStrategy.IMPUTE_OUTCOME_MEAN:
        return impute_outcome_mean(X, Y)
    elif missing_data_strategy == MissingDataStrategy.IGNORE_NANS:
        return X, None
    else:
        raise ValueError(f"Invalid missing data strategy: {missing_data_strategy}")


def impute_with(X: torch.Tensor, impute_using: torch.Tensor) -> torch.Tensor:
    """
    Impute missing entries in X with values from another tensor.

    Args:
        X (torch.Tensor): Tensor with NaN values to be replaced.
        impute_using (torch.Tensor): Tensor providing replacement values wherever X is NaN.

    Returns:
        torch.Tensor: Tensor with NaNs in X replaced by corresponding values from impute_using.
    """
    return torch.where(torch.isnan(X), impute_using, X)



import logging
from typing import Optional

import torch

from womac.config import (
    FitMethod,
    MFSMethod,
    ModelConfig,
    ReferencePoolConfig,
    Target,
    TensorLike,
    WomacConfig,
    WomacData,
    WomacResult,
)
from womac.model import AverageModel, LinearModel
from womac.preprocess import (
    _featurize_and_center,
    _missingness_features,
)
from womac.reference_pool import _build_base_reference_pool, _marginal_feature_screening
from womac.utils import (
    _apply_logit_transform,
    _cast_to_tensor,
    _get_indices_without_min_responses,
    impute_data,
    impute_with,
    jackknife_expansion,
    score_mse,
)

logger = logging.getLogger(__name__)

if torch.backends.mps.is_available():
    DEVICE = "mps"
elif torch.cuda.is_available():
    DEVICE = "cuda"
else:
    DEVICE = "cpu"


class Womac:
    def __init__(
        self,
        min_responses_per_task: int = 2,
        min_responses_per_expert: int = 2,
        device: str = DEVICE,
        seed: Optional[int] = None,
    ):
        """
        Initializes the WOMAC class.

        Args:
            min_responses_per_task (int, optional): The minimum number of responses required per task. Defaults to 2.
            min_responses_per_expert (int, optional): The minimum number of responses required per expert. Defaults to 2.
            device (str, optional): The device to be used (e.g., 'cpu' or 'cuda' or 'mps'). Defaults to DEVICE.
            seed (int | None, optional): Random seed for reproducibility. If provided, sets torch seed.
        """

        self.min_responses_per_task = min_responses_per_task
        self.min_responses_per_expert = min_responses_per_expert
        self.device = torch.device(device)
        if seed is not None:
            torch.manual_seed(seed)

        logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")

    def select_reference_pool(
        self,
        X: torch.Tensor,
        Y: torch.Tensor,
        reference_pool_config: ReferencePoolConfig,
    ) -> torch.Tensor:
        """
        Selects a pool of reference experts for each task-expert pair.

        Builds a base reference mask by applying basic configured filters in
        reference_pool_config, then uses marginal feature selection (MFS) and
        refines the mask based on reference_pool_config.mfs_config.

        Mask output dim is (dim1, dim2, n), where:
            - dim1: 1 if reference_pool_config.jackknife_row is False, m otherwise.
            - dim2: 1 if reference_pool_config.jackknife_col is False, n otherwise.
            - n: number of experts.

        Args:
            X (torch.Tensor): Prediction matrix of shape (m, n), where
                m is the number of tasks and n is the number of experts.
                May contain NaNs.
            Y (torch.Tensor): Outcome tensor of shape (m,) or (m, 1). Must have
                the same number of rows (m) as X and contain no NaNs.
            reference_pool_config (ReferencePoolConfig): Configuration object
                specifying the reference pool selection parameters.

        Returns:
            torch.BoolTensor: A boolean mask of shape (dim1, dim2, n), where
                mask[i, j, k] == True indicates that expert k is selected as
                a reference for [task i, expert j] (may be broadcast).

        Raises:
            ValueError: If X or Y fail validation in `_validate_input_data`, or if
                after applying MFS no references remain for any (i, j) pair.
        """
        # Validate inputs
        X_copy, Y_copy = _cast_to_tensor(X, device=self.device), _cast_to_tensor(Y, device=self.device)
        self._validate_input_data(X_copy, Y_copy)

        # build base reference pool
        rows_without_min_resp, cols_without_min_resp = _get_indices_without_min_responses(
            X_copy,
            self.min_responses_per_task,
            self.min_responses_per_expert,
        )

        reference_pool = _build_base_reference_pool(
            X_copy,
            reference_pool_config,
            cols_without_min_resp,
            device=self.device,
        )  # (dim1, dim2, n)

        # if no mfs_config is specified, return default mask
        if reference_pool_config.mfs_config is None:
            return reference_pool

        # score MSE for each (dim1, dim2)
        Y_copy[rows_without_min_resp] = torch.nan  # stops selected Y_i from being used in MSE calculation
        errors = score_mse(
            X_copy,
            Y_copy,
            missing_data_strategy=reference_pool_config.missing_data_strategy,
            jackknife_row=reference_pool_config.jackknife_row,
            apply_logit_transform=False,  # scoring against outcome Y, so no logit transform is applied
        )  # (dim1, n)

        # apply marginal feature screening
        reference_pool = _marginal_feature_screening(reference_pool, errors, reference_pool_config.mfs_config)
        return reference_pool  # (dim1, dim2, n)

    def get_reference_matrix(
        self,
        X: torch.Tensor,
        Y: torch.Tensor,
        womac_config: WomacConfig,
    ) -> torch.Tensor:
        """Compute the WOMAC reference for each [task, expert] pair.

        This method performs the following steps:
        1. Selects a reference-pool mask for each task and expert based on
           minimal-response filters and optional marginal feature selection.
        2. Preprocesses the data (logit transform, missing-data imputation,
           centering, and feature construction) as configured.
        3. Fits a predictive model (uniform average, linear regression, or
           logistic regression) for each hold-out [task, expert] split using
           the reference pool.
        4. Generates predictions for the held-out entries, producing the
           reference matrix. NOTE: If target is binary and feature_config.logit_feats
           is True: then if `womac_config.feature_config.score_as_probs` is False
           the matrix is in logit space; if True, matrix is in probabilities.

        Args:
            X (torch.Tensor): Input prediction matrix of shape (m, n),
                where m is the number of tasks and n is the number of experts.
            Y (torch.Tensor): Outcome tensor of shape (m,) or (m, 1).
            womac_config (WomacConfig, optional): Configuration object
                specifying reference-pool selection, preprocessing, and
                modeling parameters. Defaults to a new WomacConfig.

        Returns:
            torch.Tensor: Reference matrix of shape (m, dim2), where
                dim2 equals the second dimension of the reference-pool mask:
                - If `womac_config.reference_pool_config.jackknife_col` is True,
                  dim2 == n (one prediction per expert).
                - If False, dim2 == 1 (a single aggregated reference per task),
                  which can be broadcast against an (m, n) matrix.

        Raises:
            ValueError: If X or Y fail validation, or if any (task, expert)
                pair has no valid references.
        """
        X, Y = _cast_to_tensor(X, device=self.device), _cast_to_tensor(Y, device=self.device)
        self._validate_input_data(X, Y)

        mask = self.select_reference_pool(
            X,
            reference_pool_config=womac_config.reference_pool_config,
            Y=Y,
        )  # (dim1, dim2, n)
        womac_data = self._preprocess_data(
            X,
            Y,
            mask=mask,
            womac_config=womac_config,
        )  # (dim1, dim2, k, l)

        self._fit(
            womac_data,
            mask=mask,
            target=womac_config.target,
            logit_feats=womac_config.feature_config.logit_feats,
            model_config=womac_config.model_config,
        )
        return self._predict(
            womac_data,
            sigmoid_transform=(womac_config.feature_config.logit_feats and womac_config.feature_config.score_as_probs),
        ).squeeze(-1)  # (m, dim2, 1, l) -> (m, dim2)

    def tune_mfs(self, X: TensorLike, Y: TensorLike, womac_config: WomacConfig) -> float:
        """
        Tunes Marginal Feature Screening (MFS) percentile cutoff to minimize the Mean Squared Error (MSE)
        between the reference matrix and the target tensor `Y`. This method iteratively sweeps through
        a range of cutoff values, computes the outcome MSE for each cutoff, and selects the best cutoff
        that results in the lowest MSE.

        Args:
            X (TensorLike): Input tensor representing the features or data points. It is cast to a tensor
            and validated before processing.
            Y (TensorLike): Target tensor representing the expected outcomes or labels. It is cast to a
            tensor and validated before processing.
            womac_config (WomacConfig): Configuration object containing parameters for the Womac model,
            including the reference pool configuration and MFS tuner settings.

        Returns:
            float: The optimal cutoff value that minimizes the outcome MSE.

        Raises:
            ValueError: If the input tensors `X` and `Y` fail validation checks.
        """
        X, Y = _cast_to_tensor(X, device=self.device), _cast_to_tensor(Y, device=self.device)
        self._validate_input_data(X, Y)

        percentile_scores = []
        start = womac_config.reference_pool_config.mfs_config.start_sweep
        end = womac_config.reference_pool_config.mfs_config.end_sweep
        steps = womac_config.reference_pool_config.mfs_config.step_size

        for c in torch.arange(start, end, steps):
            womac_config.reference_pool_config.mfs_config.mfs_cutoff = c
            ref_mat = self.get_reference_matrix(X, Y, womac_config=womac_config)  # (m, dim2)
            outcome_score = torch.mean((ref_mat - Y) ** 2).cpu().float()
            percentile_scores.append((c, outcome_score))
            logging.info(f"Cutoff {c:.2f} - Outcome MSE: {outcome_score:.4f}")

        # pick the best cutoff based on outcome MSE
        best_cutoff, _ = min(percentile_scores, key=lambda x: x[1])
        logging.info(
            f"Best MFS cutoff: {best_cutoff:.2f} with outcome MSE {min(percentile_scores, key=lambda x: x[1])[1]:.4f}"
        )
        return best_cutoff

    def score_competition(
        self,
        X: TensorLike,
        Y: TensorLike,
        womac_config: WomacConfig,
    ) -> WomacResult:
        """
        Run the full WOMAC algorithm: compute processed predictions and reference values,
        and rank experts by the MSE of their predictions against the reference values and outcomes.

        This method performs the following steps:
          1. Optionally tunes the MFS cutoff if AUTO is selected, updating womac_config in-place.
          2. Imputes missing data from predictions (and applies logit transform if configured), producing
             processed predictions for each task-expert pair (m, n).
          3. Builds a reference matrix for each task/expert via get_reference_matrix (m, dim2),
             where dim2 is 1 or n depending on womac_config.reference_pool_config.jackknife_col.
          4. Computes mean squared errors (MSE) between processed X and reference values (WOMAC scores).
          5. Computes MSE between processed X and Y (outcome scores).
          6. Ranks experts by their WOMAC and outcome MSE scores.

        Args:
            X (TensorLike): Prediction matrix of shape (m, n), where
                m is the number of tasks and n is the number of experts.
            Y (TensorLike): Outcome vector of shape (m,) or matrix of shape (m, 1).
            womac_config (WomacConfig): Configuration object controlling
                reference pool selection, preprocessing options (e.g., logit transform,
                missing-data strategy), and model fitting.

        Returns:
            WomacResult: A result object with fields:
                womac_ranked_indices (torch.Tensor): Expert indices ordered by ascending WOMAC MSE, shape (n,).
                womac_ranked_scores (torch.Tensor): Sorted WOMAC MSE values, shape (n,).
                womac_scores (torch.Tensor): Raw WOMAC MSE for each expert before sorting, shape (n,).
                outcome_ranked_indices (torch.Tensor): Expert indices ordered by ascending outcome MSE, shape (n,).
                outcome_ranked_scores (torch.Tensor): Sorted outcome MSE values, shape (n,).
                outcome_scores (torch.Tensor): Raw outcome MSE for each expert before sorting, shape (n,).
                reference_matrix (torch.Tensor): Reference predictions used for scoring,
                    shape (m, dim2), where dim2 is 1 or n depending on
                    womac_config.reference_pool_config.jackknife_col.
                fitted_model: The fitted model result object.

        Raises:
            ValueError: If X or Y fail validation (wrong shape, insufficient non-NaNs, etc.).
        """
        X, Y = _cast_to_tensor(X, device=self.device), _cast_to_tensor(Y, device=self.device)
        self._validate_input_data(X, Y)

        # 1. Run AUTO MFS if configured. NOTE: this SETS the mfs_cutoff in womac_config.reference_pool_config.mfs_config
        if womac_config.reference_pool_config.mfs_config.mfs_method == MFSMethod.AUTO:
            womac_config.reference_pool_config.mfs_config.mfs_cutoff = self.tune_mfs(X, Y, womac_config=womac_config)

        # 2. Get processed predictions and reference matrix
        reference_matrix = self.get_reference_matrix(X, Y, womac_config=womac_config)  # (m, dim2)

        # 3. Score
        womac_scores = score_mse(
            X,
            reference_matrix,
            womac_config.feature_config.missing_data_strategy,
            jackknife_row=False,
            apply_logit_transform=(
                womac_config.feature_config.logit_feats and not womac_config.feature_config.score_as_probs
            ),
        ).squeeze(0)  # (n,)
        outcome_scores = score_mse(
            X,
            Y,
            womac_config.feature_config.missing_data_strategy,
            jackknife_row=False,
            apply_logit_transform=False,
        ).squeeze(0)  # (n,)

        # 4. Rank
        womac_ranked_indices = torch.argsort(womac_scores)
        womac_ranked_scores = womac_scores[womac_ranked_indices]
        outcome_ranked_indices = torch.argsort(outcome_scores)
        outcome_ranked_scores = outcome_scores[outcome_ranked_indices]

        return WomacResult(
            womac_ranked_indices=womac_ranked_indices.cpu(),
            womac_ranked_scores=womac_ranked_scores.cpu(),
            womac_scores=womac_scores.cpu(),
            outcome_ranked_indices=outcome_ranked_indices.cpu(),
            outcome_ranked_scores=outcome_ranked_scores.cpu(),
            outcome_scores=outcome_scores.cpu(),
            reference_matrix=reference_matrix.cpu(),
            fitted_model=self.model.model_result,
        )

    def _validate_input_data(self, X: torch.Tensor, Y: Optional[torch.Tensor]) -> None:
        """Validates the input data tensors X and optional Y.

        This method ensures that:
          * X is a 2D, non-empty tensor with at least two rows and two columns.
          * Every row in X has at least two non-NaN values, and at least two rows
            have ≥ self.min_responses_per_task non-NaN entries.
          * Every column in X has at least two non-NaN values, and at least two
            columns have ≥ self.min_responses_per_expert non-NaN entries.
          * If provided, Y is a 1D or 2D, non-empty tensor with at least two rows,
            contains no NaNs, and has the same number of rows as X.

        Args:
            X (torch.Tensor): A tensor of shape (tasks, experts) containing
                model predictions, possibly with NaNs.
            Y (Optional[torch.Tensor]): A tensor of shape (tasks,) or
                (tasks, 1) containing target values, or None.

        Raises:
            ValueError: If X is not 2D, is empty, has fewer than two rows or
                columns, has rows or columns with fewer than two non-NaN entries,
                or if fewer than two rows/columns meet the respective
                min_responses_per_task/expert thresholds.
            ValueError: If Y is provided but is empty, has invalid dimensions,
                contains NaNs, has fewer than two rows, or its row count
                does not match X.
        """

        # Checks on X
        if X.ndim != 2:
            raise ValueError(f"X must be a 2D array, but got {X.ndim}D.")
        if X.numel() == 0:
            raise ValueError("X must not be empty.")
        if X.shape[0] < 2:
            raise ValueError("Prediction matrix X must have at least two rows.")
        if X.shape[1] < 2:
            raise ValueError("Prediction matrix X must have at least two columns.")

        row_not_nan_counts = torch.sum(~torch.isnan(X), dim=1)
        if torch.any(row_not_nan_counts < 2):
            problematic_rows = torch.where(row_not_nan_counts < 2)[0]
            raise ValueError(f"Every row must have at least 2 non-NaN values. Problematic rows: {problematic_rows}")
        num_rows_with_min_responses = torch.sum(row_not_nan_counts >= self.min_responses_per_task)
        if num_rows_with_min_responses < 2:
            raise ValueError(
                f"Not enough rows with at least {self.min_responses_per_task} non-NaN values. "
                f"Only {num_rows_with_min_responses} rows have enough data."
            )

        col_not_nan_counts = torch.sum(~torch.isnan(X), dim=0)
        if torch.any(col_not_nan_counts < 2):
            problematic_cols = torch.where(col_not_nan_counts < 2)[0]
            raise ValueError(
                f"Every column must have at least 2 non-NaN values. Problematic columns: {problematic_cols}"
            )
        num_cols_with_min_responses = torch.sum(col_not_nan_counts >= self.min_responses_per_expert)
        if num_cols_with_min_responses < 2:
            raise ValueError(
                f"Not enough columns with at least {self.min_responses_per_expert} non-NaN values. "
                f"Only {num_cols_with_min_responses} columns have enough data."
            )

        # Checks on Y
        if Y is not None:
            if Y.ndim > 2:
                raise ValueError(f"Y must be a 1D or 2D array, but got {Y.ndim}D.")
            if Y.numel() == 0:
                raise ValueError("Y must not be empty.")
            if Y.shape[0] < 2:
                raise ValueError("Target vector Y must have at least two rows.")
            if torch.any(torch.isnan(Y)):
                raise ValueError("Y must not contain NaN values.")

            # Checks on both X and Y
            if X.shape[0] != Y.shape[0]:
                raise ValueError(
                    f"Y must have the same number of rows as X. Y has {Y.shape[0]} rows, but X has {X.shape[0]} rows."
                )

    def _preprocess_data(
        self,
        X: torch.Tensor,
        Y: torch.Tensor,
        mask: torch.Tensor,
        womac_config: WomacConfig,
    ) -> WomacData:
        """
        Preprocess input features and labels for the WOMAC model.

        This method applies, if requested, imputation of missing data according to the
        configured strategy (with logit transform if requested), jackknife expansion of
        the reference pool, centering of selected features, and extraction of missingness
        indicators.

        Args:
            X (torch.Tensor):
                Raw input feature tensor of shape (m, n), where m is the number
                of samples and n is the number of features.
            Y (torch.Tensor):
                Targets tensor of shape (m,) or (m, 1).
            mask (torch.Tensor):
                Boolean mask tensor of shape (dim1, dim2, n) indicating which
                experts can be selected as references for each [task, expert] pair.
            womac_config (WomacConfig): Configuration object containing settings for
                preprocessing,

        Returns:
            WomacData:
                A data container with the following attributes:
                - train_preds (torch.Tensor or None):
                    Preprocessed training predictors.
                    • If fit_method == UNIFORM: None
                    • If non-uniform & jackknife_row == False: shape (1, dim2, m, l)
                    • If non-uniform & jackknife_row == True: shape (m, dim2, m-1, l)
                    l is 1 or n depending on whether feature_config.pred_features is
                    FeaturizeAs.AGGREGATE or FeaturizeAs.SEPARATE. dim2 depends on
                    womac_config.reference_pool_config.jackknife_col.
                - train_missingness (torch.Tensor or None):
                    Missingness indicators for training data, same shape as train_preds.
                    l is 1 or n depending on whether feature_config.missingness_features is
                    FeaturizeAs.AGGREGATE or FeaturizeAs.SEPARATE.
                - train_labels (torch.Tensor or None):
                    Training labels aligned with train_preds.
                    • If fit_method == UNIFORM: None
                    • If non-uniform & jackknife_row == False: shape (1, 1, m)
                    • If non-uniform & jackknife_row == True: shape (m, 1, m-1)
                - test_preds (torch.Tensor):
                    Preprocessed test predictors.
                    • shape (m, 1, n)
                - test_missingness (torch.Tensor or None):
                    Missingness indicators for test data, same shape as test_preds.
                - test_labels (torch.Tensor):
                    Test labels aligned with test_preds.
                    • If reference_pool_config.jackknife_col == True: shape (m, n)
                    • If reference_pool_config.jackknife_col == False: shape (m, 1, 1)
        """

        # Initialize
        X_train, X_train_nans, X_test, X_test_nans, Y_train = None, None, None, None, None

        if womac_config.model_config.fit_method == FitMethod.UNIFORM:
            # 1A. Impute missing data
            X_test = impute_data(
                X,
                missing_data_strategy=womac_config.feature_config.missing_data_strategy,
                Y=Y,
            )[0].unsqueeze(1)  # (m, 1, n)

            # 2A. Apply logit transform if configured
            X_test = _apply_logit_transform(X_test, womac_config.feature_config.logit_feats)  # (m, 1, n)
        else:
            # 1B. Impute missing data
            if womac_config.reference_pool_config.jackknife_row:
                X_expanded = jackknife_expansion(X)  # (m, m-1, n)
                X_train, col_mean = impute_data(
                    X_expanded,
                    missing_data_strategy=womac_config.feature_config.missing_data_strategy,
                    Y=Y,
                )  # (m, m-1, n), (m, 1, n)
                X_test = impute_with(X=X.unsqueeze(1), impute_using=col_mean)  # (m, 1, n)
                Y_train = jackknife_expansion(Y).transpose(1, 2)  # (m, 1, m-1)
            else:
                X_train, col_mean = impute_data(
                    X, missing_data_strategy=womac_config.feature_config.missing_data_strategy, Y=Y
                )  # (m, n), (1, n)
                X_test = impute_with(X, col_mean)  # (m, n)
                X_train = X_train.unsqueeze(0)  # (1, m, n)
                X_test = X_test.unsqueeze(-2)  # (m, 1, n)
                Y_train = Y.view(1, -1).unsqueeze(0)  # (1, 1, m)

            # 2B. Apply logit transform if configured
            X_train = _apply_logit_transform(X_train, womac_config.feature_config.logit_feats)
            X_test = _apply_logit_transform(X_test, womac_config.feature_config.logit_feats)

            # 3. Standardize data by centering (dim1, dim2, k, l), (m, dim2, 1, l)
            X_train, X_test = _featurize_and_center(X_train, X_test, mask, womac_config.feature_config.pred_features)

            # 4. Missingness features (dim1, dim2, k, l), (m, dim2, 1, l)
            X_train_nans, X_test_nans = _missingness_features(
                X,
                mask,
                womac_config.feature_config.missingness_features,
                womac_config.reference_pool_config.jackknife_row,
            )

        if womac_config.reference_pool_config.jackknife_col:
            Y_test = Y.expand(-1, X.shape[1]).unsqueeze(-1)  # (m, n)
        else:
            Y_test = Y.unsqueeze(-1)  # (m, 1, 1)

        return WomacData(
            train_preds=X_train,
            train_missingness=X_train_nans,
            train_labels=Y_train,
            test_preds=X_test,
            test_missingness=X_test_nans,
            test_labels=Y_test,
        )

    def _fit(
        self,
        womac_data: WomacData,
        mask: torch.Tensor,
        target: Target,
        logit_feats: bool,
        model_config: ModelConfig,
    ) -> None:
        """
        Fit a WOMAC reference model according to the specified method.

        Depending on `fit_method`, this will either create an average-based model
        (UNIFORM) or a parametric model (linear or logistic regression) and train it
        on the preprocessed data.

        Args:
            womac_data (WomacData):
                Container holding preprocessed training data.
            mask (torch.Tensor):
                Boolean reference-pool mask of shape (dim1, dim2, n)
                used for uniform weighting when `fit_method` is UNIFORM.
                dim1 and dim2 are determined by jackknife_row and jackknife_col
                in the reference pool configuration.
            logit_feats (bool):
                Whether the data inputs are logit transformed (set by feature_config.logit_feats).
            model_config (ModelConfig):
                Configuration for the regression models. Passed as keyword arguments
                to the model's `.train()` method when using non-uniform fitting.
        """
        if model_config.fit_method == FitMethod.UNIFORM:
            self.model = AverageModel(target=target, logit_feats=logit_feats)
            self.model.fit(mask=mask, val_data=womac_data.test_data, val_labels=womac_data.test_labels)
        else:
            self.model = LinearModel(
                dim1=womac_data.batch_dim1,  # dim1
                dim2=womac_data.batch_dim2,  # dim2
                input_dim=womac_data.feature_dim,  # l
                target=target,
                logit_feats=logit_feats,
                device=self.device,
            )
            self.model.train(
                train_data=womac_data.train_data,
                train_labels=womac_data.train_labels,
                **model_config.kwargs_as_dict(),  # pass other model
                val_data=womac_data.test_data,
                val_labels=womac_data.test_labels,
            )

    def _predict(self, womac_data: WomacData, sigmoid_transform: bool) -> torch.Tensor:
        """Generate reference predictions for the test split.

        Uses the trained model to predict on the preprocessed test data stored
        in the WomacData container.

        Args:
            womac_data (WomacData): Container holding preprocessed tensors.
            sigmoid_transform (bool): Whether to apply a sigmoid transformation
                to the model's output. If True, the model's `predict_proba` method
                is used; otherwise, `predict` is used.

        Returns:
            torch.Tensor: Predicted reference values with shape (m, dim2, 1).
        """
        if sigmoid_transform:
            return self.model.predict_proba(womac_data.test_data)  # (m, dim2, 1)
        else:
            return self.model.predict(womac_data.test_data)  # (m, dim2, 1)



try:
    from ._version import __version__
except Exception:
    __version__ = "0+unknown"


from .cli import main
from .config import (
    FeatureConfig,
    FeaturizeAs,
    FitMethod,
    MFSConfig,
    MFSMethod,
    MissingDataStrategy,
    ModelConfig,
    ModelResult,
    ReferencePoolConfig,
    Target,
    WomacConfig,
    WomacData,
    WomacResult,
)

# public API
from .womac import Womac

__all__ = [
    "FeatureConfig",
    "FeaturizeAs",
    "FitMethod",
    "main",
    "MFSConfig",
    "MFSMethod",
    "MissingDataStrategy",
    "ModelConfig",
    "ModelResult",
    "ReferencePoolConfig",
    "Target",
    "Womac",
    "WomacConfig",
    "WomacData",
    "WomacResult",
    "__version__",
]
